""" dmt/metrics/medpy_metrics.py
General metrics for nd-images (most functionality for binary image evaluation).
Source: http://loli.github.io/medpy/metric.html

Metrics (function name):
  Note: all metric functions take numpy arrays.
  > Metrics for 2 binary arrays.
    - Dice (dc)
    - Jaccard (jc)
    - Precision (precision)
    - Recall (recall, true_positive_rate)
    - Sensitivity (sensitivity, true_negative_rate)
    - Specificity (specificity)
    - Positive Predictive Value (positive_predictive_value)
    - Hausdorff Distance (hd) 
    - Hausdorff Distance 95th percentile (hd95)
    - Average Symmetric Surface Distance (assd)
    - Average Surface Distance (asd)
    - Relative Absolute Difference (ravd)
  > Metrics for comparison of binary array sequences
    - Volume Correlation (volume_correlation)
    - Volume Change Correlation (volume_change_correlation)
  > Metrics to compare 2 sets of binary arrays
    - Object Average Symmetric Surface Distance (obj_assd)
    - Object Average Surface Distance (obj_asd)
    - Object False Positive Rate (obj_fpr)
    - Object True Positive Rate (obj_tpr)
Helpers: 
    - Distinct Binary Object Correspondences
    - Surface Distances
    - Combine Windows
"""


import numpy
from scipy.ndimage import _ni_support
from scipy.ndimage.morphology import distance_transform_edt, binary_erosion,\
    generate_binary_structure
from scipy.ndimage.measurements import label, find_objects
from scipy.stats import pearsonr

# own modules

# code
def dc(result, reference):
    r"""
    Dice coefficient
    
    Computes the Dice coefficient (also known as Sorensen index) between the binary
    objects in two images.
    
    The metric is defined as
    
    .. math::
        
        DC=\frac{2|A\cap B|}{|A|+|B|}
        
    , where :math:`A` is the first and :math:`B` the second set of samples (here: binary objects).
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    
    Returns
    -------
    dc : float
        The Dice coefficient between the object(s) in ```result``` and the
        object(s) in ```reference```. It ranges from 0 (no overlap) to 1 (perfect overlap).
        
    Notes
    -----
    This is a real metric. The binary images can therefore be supplied in any order.
    """
    result = numpy.atleast_1d(result.astype(numpy.bool))
    reference = numpy.atleast_1d(reference.astype(numpy.bool))
    
    intersection = numpy.count_nonzero(result & reference)
    
    size_i1 = numpy.count_nonzero(result)
    size_i2 = numpy.count_nonzero(reference)
    
    try:
        dc = 2. * intersection / float(size_i1 + size_i2)
    except ZeroDivisionError:
        dc = 0.0
    
    return dc

def jc(result, reference):
    """
    Jaccard coefficient
    
    Computes the Jaccard coefficient between the binary objects in two images.
    
    Parameters
    ----------
    result: array_like
            Input data containing objects. Can be any type but will be converted
            into binary: background where 0, object everywhere else.
    reference: array_like
            Input data containing objects. Can be any type but will be converted
            into binary: background where 0, object everywhere else.
    Returns
    -------
    jc: float
        The Jaccard coefficient between the object(s) in `result` and the
        object(s) in `reference`. It ranges from 0 (no overlap) to 1 (perfect overlap).
    
    Notes
    -----
    This is a real metric. The binary images can therefore be supplied in any order.
    """
    result = numpy.atleast_1d(result.astype(numpy.bool))
    reference = numpy.atleast_1d(reference.astype(numpy.bool))
    
    intersection = numpy.count_nonzero(result & reference)
    union = numpy.count_nonzero(result | reference)
    
    jc = float(intersection) / float(union)
    
    return jc

def precision(result, reference):
    """
    Precison.
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    
    Returns
    -------
    precision : float
        The precision between two binary datasets, here mostly binary objects in images,
        which is defined as the fraction of retrieved instances that are relevant. The
        precision is not symmetric.
    
    See also
    --------
    :func:`recall`
    
    Notes
    -----
    Not symmetric. The inverse of the precision is :func:`recall`.
    High precision means that an algorithm returned substantially more relevant results than irrelevant.
    
    References
    ----------
    .. [1] http://en.wikipedia.org/wiki/Precision_and_recall
    .. [2] http://en.wikipedia.org/wiki/Confusion_matrix#Table_of_confusion
    """
    result = numpy.atleast_1d(result.astype(numpy.bool))
    reference = numpy.atleast_1d(reference.astype(numpy.bool))
        
    tp = numpy.count_nonzero(result & reference)
    fp = numpy.count_nonzero(result & ~reference)
    
    try:
        precision = tp / float(tp + fp)
    except ZeroDivisionError:
        precision = 0.0
    
    return precision

def recall(result, reference):
    """
    Recall.
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    
    Returns
    -------
    recall : float
        The recall between two binary datasets, here mostly binary objects in images,
        which is defined as the fraction of relevant instances that are retrieved. The
        recall is not symmetric.
    
    See also
    --------
    :func:`precision`
    
    Notes
    -----
    Not symmetric. The inverse of the recall is :func:`precision`.
    High recall means that an algorithm returned most of the relevant results.
    
    References
    ----------
    .. [1] http://en.wikipedia.org/wiki/Precision_and_recall
    .. [2] http://en.wikipedia.org/wiki/Confusion_matrix#Table_of_confusion
    """
    result = numpy.atleast_1d(result.astype(numpy.bool))
    reference = numpy.atleast_1d(reference.astype(numpy.bool))
        
    tp = numpy.count_nonzero(result & reference)
    fn = numpy.count_nonzero(~result & reference)

    try:
        recall = tp / float(tp + fn)
    except ZeroDivisionError:
        recall = 0.0
    
    return recall

def sensitivity(result, reference):
    """
    Sensitivity.
    Same as :func:`recall`, see there for a detailed description.
    
    See also
    --------
    :func:`specificity` 
    """
    return recall(result, reference)

def specificity(result, reference):
    """
    Specificity.
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    
    Returns
    -------
    specificity : float
        The specificity between two binary datasets, here mostly binary objects in images,
        which denotes the fraction of correctly returned negatives. The
        specificity is not symmetric.
    
    See also
    --------
    :func:`sensitivity`
    
    Notes
    -----
    Not symmetric. The completment of the specificity is :func:`sensitivity`.
    High recall means that an algorithm returned most of the irrelevant results.
    
    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Sensitivity_and_specificity
    .. [2] http://en.wikipedia.org/wiki/Confusion_matrix#Table_of_confusion
    """
    result = numpy.atleast_1d(result.astype(numpy.bool))
    reference = numpy.atleast_1d(reference.astype(numpy.bool))
       
    tn = numpy.count_nonzero(~result & ~reference)
    fp = numpy.count_nonzero(result & ~reference)

    try:
        specificity = tn / float(tn + fp)
    except ZeroDivisionError:
        specificity = 0.0
    
    return specificity


def true_negative_rate(result, reference):
    """
    True negative rate.
    Same as :func:`sensitivity`, see there for a detailed description.
    
    See also
    --------
    :func:`true_positive_rate` 
    :func:`positive_predictive_value`
    """
    return sensitivity(result, reference)


def true_positive_rate(result, reference):
    """
    True positive rate.
    Same as :func:`recall`, see there for a detailed description.
    
    See also
    --------
    :func:`positive_predictive_value` 
    :func:`true_negative_rate`
    """
    return recall(result, reference)


def positive_predictive_value(result, reference):
    """
    Positive predictive value.
    Same as :func:`precision`, see there for a detailed description.
    
    See also
    --------
    :func:`true_positive_rate`
    :func:`true_negative_rate`
    """
    return precision(result, reference)

def hd(result, reference, voxelspacing=None, connectivity=1):
    """
    Hausdorff Distance.
    
    Computes the (symmetric) Hausdorff Distance (HD) between the binary objects in two
    images. It is defined as the maximum surface distance between the objects.
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    voxelspacing : float or sequence of floats, optional
        The voxelspacing in a distance unit i.e. spacing of elements
        along each dimension. If a sequence, must be of length equal to
        the input rank; if a single number, this is used for all axes. If
        not specified, a grid spacing of unity is implied.
    connectivity : int
        The neighbourhood/connectivity considered when determining the surface
        of the binary objects. This value is passed to
        `scipy.ndimage.morphology.generate_binary_structure` and should usually be :math:`> 1`.
        Note that the connectivity influences the result in the case of the Hausdorff distance.
        
    Returns
    -------
    hd : float
        The symmetric Hausdorff Distance between the object(s) in ```result``` and the
        object(s) in ```reference```. The distance unit is the same as for the spacing of 
        elements along each dimension, which is usually given in mm.
        
    See also
    --------
    :func:`assd`
    :func:`asd`
    
    Notes
    -----
    This is a real metric. The binary images can therefore be supplied in any order.
    """
    hd1 = __surface_distances(result, reference, voxelspacing, connectivity).max()
    hd2 = __surface_distances(reference, result, voxelspacing, connectivity).max()
    hd = max(hd1, hd2)
    return hd


def hd95(result, reference, voxelspacing=None, connectivity=1):
    """
    95th percentile of the Hausdorff Distance.
    Computes the 95th percentile of the (symmetric) Hausdorff Distance (HD) between the binary objects in two
    images. Compared to the Hausdorff Distance, this metric is slightly more stable to small outliers and is
    commonly used in Biomedical Segmentation challenges.
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    voxelspacing : float or sequence of floats, optional
        The voxelspacing in a distance unit i.e. spacing of elements
        along each dimension. If a sequence, must be of length equal to
        the input rank; if a single number, this is used for all axes. If
        not specified, a grid spacing of unity is implied.
    connectivity : int
        The neighbourhood/connectivity considered when determining the surface
        of the binary objects. This value is passed to
        `scipy.ndimage.morphology.generate_binary_structure` and should usually be :math:`> 1`.
        Note that the connectivity influences the result in the case of the Hausdorff distance.
    Returns
    -------
    hd : float
        The symmetric Hausdorff Distance between the object(s) in ```result``` and the
        object(s) in ```reference```. The distance unit is the same as for the spacing of 
        elements along each dimension, which is usually given in mm.
    See also
    --------
    :func:`hd`
    Notes
    -----
    This is a real metric. The binary images can therefore be supplied in any order.
    """
    hd1 = __surface_distances(result, reference, voxelspacing, connectivity)
    hd2 = __surface_distances(reference, result, voxelspacing, connectivity)
    hd95 = numpy.percentile(numpy.hstack((hd1, hd2)), 95)
    return hd95


def assd(result, reference, voxelspacing=None, connectivity=1):
    """
    Average symmetric surface distance.
    
    Computes the average symmetric surface distance (ASD) between the binary objects in
    two images.
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    voxelspacing : float or sequence of floats, optional
        The voxelspacing in a distance unit i.e. spacing of elements
        along each dimension. If a sequence, must be of length equal to
        the input rank; if a single number, this is used for all axes. If
        not specified, a grid spacing of unity is implied.
    connectivity : int
        The neighbourhood/connectivity considered when determining the surface
        of the binary objects. This value is passed to
        `scipy.ndimage.morphology.generate_binary_structure` and should usually be :math:`> 1`.
        The decision on the connectivity is important, as it can influence the results
        strongly. If in doubt, leave it as it is.         
        
    Returns
    -------
    assd : float
        The average symmetric surface distance between the object(s) in ``result`` and the
        object(s) in ``reference``. The distance unit is the same as for the spacing of 
        elements along each dimension, which is usually given in mm.
        
    See also
    --------
    :func:`asd`
    :func:`hd`
    
    Notes
    -----
    This is a real metric, obtained by calling and averaging
    
    >>> asd(result, reference)
    
    and
    
    >>> asd(reference, result)
    
    The binary images can therefore be supplied in any order.
    """
    assd = numpy.mean( (asd(result, reference, voxelspacing, connectivity), asd(reference, result, voxelspacing, connectivity)) )
    return assd


def asd(result, reference, voxelspacing=None, connectivity=1):
    """
    Average surface distance metric.
    
    Computes the average surface distance (ASD) between the binary objects in two images.
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    voxelspacing : float or sequence of floats, optional
        The voxelspacing in a distance unit i.e. spacing of elements
        along each dimension. If a sequence, must be of length equal to
        the input rank; if a single number, this is used for all axes. If
        not specified, a grid spacing of unity is implied.
    connectivity : int
        The neighbourhood/connectivity considered when determining the surface
        of the binary objects. This value is passed to
        `scipy.ndimage.morphology.generate_binary_structure` and should usually be :math:`> 1`.
        The decision on the connectivity is important, as it can influence the results
        strongly. If in doubt, leave it as it is.
    
    Returns
    -------
    asd : float
        The average surface distance between the object(s) in ``result`` and the
        object(s) in ``reference``. The distance unit is the same as for the spacing
        of elements along each dimension, which is usually given in mm.
        
    See also
    --------
    :func:`assd`
    :func:`hd`
    
    
    Notes
    -----
    This is not a real metric, as it is directed. See `assd` for a real metric of this.
    
    The method is implemented making use of distance images and simple binary morphology
    to achieve high computational speed.
    
    Examples
    --------
    The `connectivity` determines what pixels/voxels are considered the surface of a
    binary object. Take the following binary image showing a cross
    
    >>> from scipy.ndimage.morphology import generate_binary_structure
    >>> cross = generate_binary_structure(2, 1)
    array([[0, 1, 0],
           [1, 1, 1],
           [0, 1, 0]])
           
    With `connectivity` set to `1` a 4-neighbourhood is considered when determining the
    object surface, resulting in the surface
    
    .. code-block:: python
    
        array([[0, 1, 0],
               [1, 0, 1],
               [0, 1, 0]])
           
    Changing `connectivity` to `2`, a 8-neighbourhood is considered and we get:
    
    .. code-block:: python
    
        array([[0, 1, 0],
               [1, 1, 1],
               [0, 1, 0]])
           
    , as a diagonal connection does no longer qualifies as valid object surface.
    
    This influences the  results `asd` returns. Imagine we want to compute the surface
    distance of our cross to a cube-like object:
    
    >>> cube = generate_binary_structure(2, 1)
    array([[1, 1, 1],
           [1, 1, 1],
           [1, 1, 1]])
           
    , which surface is, independent of the `connectivity` value set, always
    
    .. code-block:: python
    
        array([[1, 1, 1],
               [1, 0, 1],
               [1, 1, 1]])
           
    Using a `connectivity` of `1` we get
    
    >>> asd(cross, cube, connectivity=1)
    0.0
    
    while a value of `2` returns us
    
    >>> asd(cross, cube, connectivity=2)
    0.20000000000000001
    
    due to the center of the cross being considered surface as well.
    
    """
    sds = __surface_distances(result, reference, voxelspacing, connectivity)
    asd = sds.mean()
    return asd


def ravd(result, reference):
    """
    Relative absolute volume difference.
    
    Compute the relative absolute volume difference between the (joined) binary objects
    in the two images.
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
        
    Returns
    -------
    ravd : float
        The relative absolute volume difference between the object(s) in ``result``
        and the object(s) in ``reference``. This is a percentage value in the range
        :math:`[-1.0, +inf]` for which a :math:`0` denotes an ideal score.
        
    Raises
    ------
    RuntimeError
        If the reference object is empty.
        
    See also
    --------
    :func:`dc`
    :func:`precision`
    :func:`recall`
    
    Notes
    -----
    This is not a real metric, as it is directed. Negative values denote a smaller
    and positive values a larger volume than the reference.
    This implementation does not check, whether the two supplied arrays are of the same
    size.
    
    Examples
    --------
    Considering the following inputs
    
    >>> import numpy
    >>> arr1 = numpy.asarray([[0,1,0],[1,1,1],[0,1,0]])
    >>> arr1
    array([[0, 1, 0],
           [1, 1, 1],
           [0, 1, 0]])
    >>> arr2 = numpy.asarray([[0,1,0],[1,0,1],[0,1,0]])
    >>> arr2
    array([[0, 1, 0],
           [1, 0, 1],
           [0, 1, 0]])
           
    comparing `arr1` to `arr2` we get
    
    >>> ravd(arr1, arr2)
    -0.2
    
    and reversing the inputs the directivness of the metric becomes evident
    
    >>> ravd(arr2, arr1)
    0.25
    
    It is important to keep in mind that a perfect score of `0` does not mean that the
    binary objects fit exactely, as only the volumes are compared:
    
    >>> arr1 = numpy.asarray([1,0,0])
    >>> arr2 = numpy.asarray([0,0,1])
    >>> ravd(arr1, arr2)
    0.0
    
    """
    result = numpy.atleast_1d(result.astype(numpy.bool))
    reference = numpy.atleast_1d(reference.astype(numpy.bool))
        
    vol1 = numpy.count_nonzero(result)
    vol2 = numpy.count_nonzero(reference)
    
    if 0 == vol2:
        raise RuntimeError('The second supplied array does not contain any binary object.')
    
    return (vol1 - vol2) / float(vol2)


def volume_correlation(results, references):
    r"""
    Volume correlation.
    
    Computes the linear correlation in binary object volume between the
    contents of the successive binary images supplied. Measured through
    the Pearson product-moment correlation coefficient. 
    
    Parameters
    ----------
    results : sequence of array_like
        Ordered list of input data containing objects. Each array_like will be
        converted into binary: background where 0, object everywhere else.
    references : sequence of array_like
        Ordered list of input data containing objects. Each array_like will be
        converted into binary: background where 0, object everywhere else.
        The order must be the same as for ``results``.
    
    Returns
    -------
    r : float
        The correlation coefficient between -1 and 1.
    p : float
        The two-side p value.
        
    """
    results = numpy.atleast_2d(numpy.array(results).astype(numpy.bool))
    references = numpy.atleast_2d(numpy.array(references).astype(numpy.bool))
    
    results_volumes = [numpy.count_nonzero(r) for r in results]
    references_volumes = [numpy.count_nonzero(r) for r in references]
    
    return pearsonr(results_volumes, references_volumes) # returns (Pearson'


def volume_change_correlation(results, references):
    r"""
    Volume change correlation.
    
    Computes the linear correlation of change in binary object volume between
    the contents of the successive binary images supplied. Measured through
    the Pearson product-moment correlation coefficient. 
    
    Parameters
    ----------
    results : sequence of array_like
        Ordered list of input data containing objects. Each array_like will be
        converted into binary: background where 0, object everywhere else.
    references : sequence of array_like
        Ordered list of input data containing objects. Each array_like will be
        converted into binary: background where 0, object everywhere else.
        The order must be the same as for ``results``.
    
    Returns
    -------
    r : float
        The correlation coefficient between -1 and 1.
    p : float
        The two-side p value.
        
    """
    results = numpy.atleast_2d(numpy.array(results).astype(numpy.bool))
    references = numpy.atleast_2d(numpy.array(references).astype(numpy.bool))
    
    results_volumes = numpy.asarray([numpy.count_nonzero(r) for r in results])
    references_volumes = numpy.asarray([numpy.count_nonzero(r) for r in references])
    
    results_volumes_changes = results_volumes[1:] - results_volumes[:-1]
    references_volumes_changes = references_volumes[1:] - references_volumes[:-1] 
    
    return pearsonr(results_volumes_changes, references_volumes_changes) # returns (Pearson's correlation coefficient, 2-tailed p-value)


def obj_assd(result, reference, voxelspacing=None, connectivity=1):
    """
    Average symmetric surface distance.
    
    Computes the average symmetric surface distance (ASSD) between the binary objects in
    two images.
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    voxelspacing : float or sequence of floats, optional
        The voxelspacing in a distance unit i.e. spacing of elements
        along each dimension. If a sequence, must be of length equal to
        the input rank; if a single number, this is used for all axes. If
        not specified, a grid spacing of unity is implied.
    connectivity : int
        The neighbourhood/connectivity considered when determining what accounts
        for a distinct binary object as well as when determining the surface
        of the binary objects. This value is passed to
        `scipy.ndimage.morphology.generate_binary_structure` and should usually be :math:`> 1`.
        The decision on the connectivity is important, as it can influence the results
        strongly. If in doubt, leave it as it is.
        
    Returns
    -------
    assd : float
        The average symmetric surface distance between all mutually existing distinct
        binary object(s) in ``result`` and ``reference``. The distance unit is the same as for
        the spacing of elements along each dimension, which is usually given in mm.
        
    See also
    --------
    :func:`obj_asd`
    
    Notes
    -----
    This is a real metric, obtained by calling and averaging
    
    >>> obj_asd(result, reference)
    
    and
    
    >>> obj_asd(reference, result)
    
    The binary images can therefore be supplied in any order.
    """
    assd = numpy.mean( (obj_asd(result, reference, voxelspacing, connectivity), obj_asd(reference, result, voxelspacing, connectivity)) )
    return assd
    
    
def obj_asd(result, reference, voxelspacing=None, connectivity=1):
    """
    Average surface distance between objects.
    
    First correspondences between distinct binary objects in reference and result are
    established. Then the average surface distance is only computed between corresponding
    objects. Correspondence is defined as unique and at least one voxel overlap.
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    voxelspacing : float or sequence of floats, optional
        The voxelspacing in a distance unit i.e. spacing of elements
        along each dimension. If a sequence, must be of length equal to
        the input rank; if a single number, this is used for all axes. If
        not specified, a grid spacing of unity is implied.
    connectivity : int
        The neighbourhood/connectivity considered when determining what accounts
        for a distinct binary object as well as when determining the surface
        of the binary objects. This value is passed to
        `scipy.ndimage.morphology.generate_binary_structure` and should usually be :math:`> 1`.
        The decision on the connectivity is important, as it can influence the results
        strongly. If in doubt, leave it as it is.
        
    Returns
    -------
    asd : float
        The average surface distance between all mutually existing distinct binary
        object(s) in ``result`` and ``reference``. The distance unit is the same as for the
        spacing of elements along each dimension, which is usually given in mm.
        
    See also
    --------
    :func:`obj_assd`
    :func:`obj_tpr`
    :func:`obj_fpr`
        
    Notes
    -----
    This is not a real metric, as it is directed. See `obj_assd` for a real metric of this.
    
    For the understanding of this metric, both the notions of connectedness and surface
    distance are essential. Please see :func:`obj_tpr` and :func:`obj_fpr` for more
    information on the first and :func:`asd` on the second.
        
    Examples
    --------
    >>> arr1 = numpy.asarray([[1,1,1],[1,1,1],[1,1,1]])
    >>> arr2 = numpy.asarray([[0,1,0],[0,1,0],[0,1,0]])
    >>> arr1
    array([[1, 1, 1],
           [1, 1, 1],
           [1, 1, 1]])
    >>> arr2
    array([[0, 1, 0],
           [0, 1, 0],
           [0, 1, 0]])
    >>> obj_asd(arr1, arr2)
    1.5
    >>> obj_asd(arr2, arr1)
    0.333333333333
    
    With the `voxelspacing` parameter, the distances between the voxels can be set for
    each dimension separately:
    
    >>> obj_asd(arr1, arr2, voxelspacing=(1,2))
    1.5
    >>> obj_asd(arr2, arr1, voxelspacing=(1,2))
    0.333333333333    
    
    More examples depicting the notion of object connectedness:
    
    >>> arr1 = numpy.asarray([[1,0,1],[1,0,0],[0,0,0]])
    >>> arr2 = numpy.asarray([[1,0,1],[1,0,0],[0,0,1]])
    >>> arr1
    array([[1, 0, 1],
           [1, 0, 0],
           [0, 0, 0]])
    >>> arr2
    array([[1, 0, 1],
           [1, 0, 0],
           [0, 0, 1]])
    >>> obj_asd(arr1, arr2)
    0.0
    >>> obj_asd(arr2, arr1)
    0.0
    
    >>> arr1 = numpy.asarray([[1,0,1],[1,0,1],[0,0,1]])
    >>> arr2 = numpy.asarray([[1,0,1],[1,0,0],[0,0,1]])
    >>> arr1
    array([[1, 0, 1],
           [1, 0, 1],
           [0, 0, 1]])
    >>> arr2
    array([[1, 0, 1],
           [1, 0, 0],
           [0, 0, 1]])
    >>> obj_asd(arr1, arr2)
    0.6
    >>> obj_asd(arr2, arr1)
    0.0
    
    Influence of `connectivity` parameter can be seen in the following example, where
    with the (default) connectivity of `1` the first array is considered to contain two
    objects, while with an increase connectivity of `2`, just one large object is
    detected.  
    
    >>> arr1 = numpy.asarray([[1,0,0],[0,1,1],[0,1,1]])
    >>> arr2 = numpy.asarray([[1,0,0],[0,0,0],[0,0,0]])
    >>> arr1
    array([[1, 0, 0],
           [0, 1, 1],
           [0, 1, 1]])
    >>> arr2
    array([[1, 0, 0],
           [0, 0, 0],
           [0, 0, 0]])
    >>> obj_asd(arr1, arr2)
    0.0
    >>> obj_asd(arr1, arr2, connectivity=2)
    1.742955328
    
    Note that the connectivity also influence the notion of what is considered an object
    surface voxels.
    """
    sds = list()
    labelmap1, labelmap2, _a, _b, mapping = __distinct_binary_object_correspondences(result, reference, connectivity)
    slicers1 = find_objects(labelmap1)
    slicers2 = find_objects(labelmap2)
    for lid2, lid1 in mapping.items():
        window = __combine_windows(slicers1[lid1 - 1], slicers2[lid2 - 1])
        object1 = labelmap1[window] == lid1
        object2 = labelmap2[window] == lid2
        sds.extend(__surface_distances(object1, object2, voxelspacing, connectivity))
    asd = numpy.mean(sds)
    return asd


def obj_fpr(result, reference, connectivity=1):
    """
    The false positive rate of distinct binary object detection.
    
    The false positive rates gives a percentage measure of how many distinct binary
    objects in the second array do not exists in the first array. A partial overlap
    (of minimum one voxel) is here considered sufficient.
    
    In cases where two distinct binary object in the second array overlap with a single
    distinct object in the first array, only one is considered to have been detected
    successfully and the other is added to the count of false positives.
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    connectivity : int
        The neighbourhood/connectivity considered when determining what accounts
        for a distinct binary object. This value is passed to
        `scipy.ndimage.morphology.generate_binary_structure` and should usually be :math:`> 1`.
        The decision on the connectivity is important, as it can influence the results
        strongly. If in doubt, leave it as it is.
        
    Returns
    -------
    tpr : float
        A percentage measure of how many distinct binary objects in ``results`` have no
        corresponding binary object in ``reference``. It has the range :math:`[0, 1]`, where a :math:`0`
        denotes an ideal score.
        
    Raises
    ------
    RuntimeError
        If the second array is empty.
    
    See also
    --------
    :func:`obj_tpr`
    
    Notes
    -----
    This is not a real metric, as it is directed. Whatever array is considered as
    reference should be passed second. A perfect score of :math:`0` tells that there are no
    distinct binary objects in the second array that do not exists also in the reference
    array, but does not reveal anything about objects in the reference array also
    existing in the second array (use :func:`obj_tpr` for this).
    
    Examples
    --------
    >>> arr2 = numpy.asarray([[1,0,0],[1,0,1],[0,0,1]])
    >>> arr1 = numpy.asarray([[0,0,1],[1,0,1],[0,0,1]])
    >>> arr2
    array([[1, 0, 0],
           [1, 0, 1],
           [0, 0, 1]])
    >>> arr1
    array([[0, 0, 1],
           [1, 0, 1],
           [0, 0, 1]])
    >>> obj_fpr(arr1, arr2)
    0.0
    >>> obj_fpr(arr2, arr1)
    0.0
    
    Example of directedness:
    
    >>> arr2 = numpy.asarray([1,0,1,0,1])
    >>> arr1 = numpy.asarray([1,0,1,0,0])
    >>> obj_fpr(arr1, arr2)
    0.0
    >>> obj_fpr(arr2, arr1)
    0.3333333333333333
    
    Examples of multiple overlap treatment:
    
    >>> arr2 = numpy.asarray([1,0,1,0,1,1,1])
    >>> arr1 = numpy.asarray([1,1,1,0,1,0,1])
    >>> obj_fpr(arr1, arr2)
    0.3333333333333333
    >>> obj_fpr(arr2, arr1)
    0.3333333333333333
    
    >>> arr2 = numpy.asarray([1,0,1,1,1,0,1])
    >>> arr1 = numpy.asarray([1,1,1,0,1,1,1])
    >>> obj_fpr(arr1, arr2)
    0.0
    >>> obj_fpr(arr2, arr1)
    0.3333333333333333
    
    >>> arr2 = numpy.asarray([[1,0,1,0,0],
                              [1,0,0,0,0],
                              [1,0,1,1,1],
                              [0,0,0,0,0],
                              [1,0,1,0,0]])
    >>> arr1 = numpy.asarray([[1,1,1,0,0],
                              [0,0,0,0,0],
                              [1,1,1,0,1],
                              [0,0,0,0,0],
                              [1,1,1,0,0]])
    >>> obj_fpr(arr1, arr2)
    0.0
    >>> obj_fpr(arr2, arr1)
    0.2    
    """
    _, _, _, n_obj_reference, mapping = __distinct_binary_object_correspondences(reference, result, connectivity)
    return (n_obj_reference - len(mapping)) / float(n_obj_reference)


def obj_tpr(result, reference, connectivity=1):
    """
    The true positive rate of distinct binary object detection.
    
    The true positive rates gives a percentage measure of how many distinct binary
    objects in the first array also exists in the second array. A partial overlap
    (of minimum one voxel) is here considered sufficient.
    
    In cases where two distinct binary object in the first array overlaps with a single
    distinct object in the second array, only one is considered to have been detected
    successfully.  
    
    Parameters
    ----------
    result : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    reference : array_like
        Input data containing objects. Can be any type but will be converted
        into binary: background where 0, object everywhere else.
    connectivity : int
        The neighbourhood/connectivity considered when determining what accounts
        for a distinct binary object. This value is passed to
        `scipy.ndimage.morphology.generate_binary_structure` and should usually be :math:`> 1`.
        The decision on the connectivity is important, as it can influence the results
        strongly. If in doubt, leave it as it is.
        
    Returns
    -------
    tpr : float
        A percentage measure of how many distinct binary objects in ``result`` also exists
        in ``reference``. It has the range :math:`[0, 1]`, where a :math:`1` denotes an ideal score.
        
    Raises
    ------
    RuntimeError
        If the reference object is empty.
    
    See also
    --------
    :func:`obj_fpr`
    
    Notes
    -----
    This is not a real metric, as it is directed. Whatever array is considered as
    reference should be passed second. A perfect score of :math:`1` tells that all distinct
    binary objects in the reference array also exist in the result array, but does not
    reveal anything about additional binary objects in the result array
    (use :func:`obj_fpr` for this).
    
    Examples
    --------
    >>> arr2 = numpy.asarray([[1,0,0],[1,0,1],[0,0,1]])
    >>> arr1 = numpy.asarray([[0,0,1],[1,0,1],[0,0,1]])
    >>> arr2
    array([[1, 0, 0],
           [1, 0, 1],
           [0, 0, 1]])
    >>> arr1
    array([[0, 0, 1],
           [1, 0, 1],
           [0, 0, 1]])
    >>> obj_tpr(arr1, arr2)
    1.0
    >>> obj_tpr(arr2, arr1)
    1.0
    
    Example of directedness:
    
    >>> arr2 = numpy.asarray([1,0,1,0,1])
    >>> arr1 = numpy.asarray([1,0,1,0,0])
    >>> obj_tpr(arr1, arr2)
    0.6666666666666666
    >>> obj_tpr(arr2, arr1)
    1.0
    
    Examples of multiple overlap treatment:
    
    >>> arr2 = numpy.asarray([1,0,1,0,1,1,1])
    >>> arr1 = numpy.asarray([1,1,1,0,1,0,1])
    >>> obj_tpr(arr1, arr2)
    0.6666666666666666
    >>> obj_tpr(arr2, arr1)
    0.6666666666666666
    
    >>> arr2 = numpy.asarray([1,0,1,1,1,0,1])
    >>> arr1 = numpy.asarray([1,1,1,0,1,1,1])
    >>> obj_tpr(arr1, arr2)
    0.6666666666666666
    >>> obj_tpr(arr2, arr1)
    1.0
    
    >>> arr2 = numpy.asarray([[1,0,1,0,0],
                              [1,0,0,0,0],
                              [1,0,1,1,1],
                              [0,0,0,0,0],
                              [1,0,1,0,0]])
    >>> arr1 = numpy.asarray([[1,1,1,0,0],
                              [0,0,0,0,0],
                              [1,1,1,0,1],
                              [0,0,0,0,0],
                              [1,1,1,0,0]])
    >>> obj_tpr(arr1, arr2)
    0.8
    >>> obj_tpr(arr2, arr1)
    1.0    
    """
    _, _, n_obj_result, _, mapping = __distinct_binary_object_correspondences(reference, result, connectivity)
    return len(mapping) / float(n_obj_result)


def __distinct_binary_object_correspondences(reference, result, connectivity=1):
    """
    Determines all distinct (where connectivity is defined by the connectivity parameter
    passed to scipy's `generate_binary_structure`) binary objects in both of the input
    parameters and returns a 1to1 mapping from the labelled objects in reference to the
    corresponding (whereas a one-voxel overlap suffices for correspondence) objects in
    result.
    
    All stems from the problem, that the relationship is non-surjective many-to-many.
    
    @return (labelmap1, labelmap2, n_lables1, n_labels2, labelmapping2to1)
    """
    result = numpy.atleast_1d(result.astype(numpy.bool))
    reference = numpy.atleast_1d(reference.astype(numpy.bool))
    
    # binary structure
    footprint = generate_binary_structure(result.ndim, connectivity)
    
    # label distinct binary objects
    labelmap1, n_obj_result = label(result, footprint)
    labelmap2, n_obj_reference = label(reference, footprint)
    
    # find all overlaps from labelmap2 to labelmap1; collect one-to-one relationships and store all one-two-many for later processing
    slicers = find_objects(labelmap2) # get windows of labelled objects
    mapping = dict() # mappings from labels in labelmap2 to corresponding object labels in labelmap1
    used_labels = set() # set to collect all already used labels from labelmap2
    one_to_many = list() # list to collect all one-to-many mappings
    for l1id, slicer in enumerate(slicers): # iterate over object in labelmap2 and their windows
        l1id += 1 # labelled objects have ids sarting from 1
        bobj = (l1id) == labelmap2[slicer] # find binary object corresponding to the label1 id in the segmentation
        l2ids = numpy.unique(labelmap1[slicer][bobj]) # extract all unique object identifiers at the corresponding positions in the reference (i.e. the mapping)
        l2ids = l2ids[0 != l2ids] # remove background identifiers (=0)
        if 1 == len(l2ids): # one-to-one mapping: if target label not already used, add to final list of object-to-object mappings and mark target label as used
            l2id = l2ids[0]
            if not l2id in used_labels:
                mapping[l1id] = l2id
                used_labels.add(l2id)
        elif 1 < len(l2ids): # one-to-many mapping: store relationship for later processing
            one_to_many.append((l1id, set(l2ids)))
            
    # process one-to-many mappings, always choosing the one with the least labelmap2 correspondences first
    while True:
        one_to_many = [(l1id, l2ids - used_labels) for l1id, l2ids in one_to_many] # remove already used ids from all sets
        one_to_many = [x for x in one_to_many if x[1]] # remove empty sets
        one_to_many = sorted(one_to_many, key=lambda x: len(x[1])) # sort by set length
        if 0 == len(one_to_many):
            break
        l2id = one_to_many[0][1].pop() # select an arbitrary target label id from the shortest set
        mapping[one_to_many[0][0]] = l2id # add to one-to-one mappings 
        used_labels.add(l2id) # mark target label as used
        one_to_many = one_to_many[1:] # delete the processed set from all sets
    
    return labelmap1, labelmap2, n_obj_result, n_obj_reference, mapping


def __surface_distances(result, reference, voxelspacing=None, connectivity=1):
    """
    The distances between the surface voxel of binary objects in result and their
    nearest partner surface voxel of a binary object in reference.
    """
    result = numpy.atleast_1d(result.astype(numpy.bool))
    reference = numpy.atleast_1d(reference.astype(numpy.bool))
    if voxelspacing is not None:
        voxelspacing = _ni_support._normalize_sequence(voxelspacing, result.ndim)
        voxelspacing = numpy.asarray(voxelspacing, dtype=numpy.float64)
        if not voxelspacing.flags.contiguous:
            voxelspacing = voxelspacing.copy()
            
    # binary structure
    footprint = generate_binary_structure(result.ndim, connectivity)
    
    # test for emptiness
    if 0 == numpy.count_nonzero(result): 
        raise RuntimeError('The first supplied array does not contain any binary object.')
    if 0 == numpy.count_nonzero(reference): 
        raise RuntimeError('The second supplied array does not contain any binary object.')    
            
    # extract only 1-pixel border line of objects
    result_border = result ^ binary_erosion(result, structure=footprint, iterations=1)
    reference_border = reference ^ binary_erosion(reference, structure=footprint, iterations=1)
    
    # compute average surface distance        
    # Note: scipys distance transform is calculated only inside the borders of the
    #       foreground objects, therefore the input has to be reversed
    dt = distance_transform_edt(~reference_border, sampling=voxelspacing)
    sds = dt[result_border]
    
    return sds


def __combine_windows(w1, w2):
    """
    Joins two windows (defined by tuple of slices) such that their maximum
    combined extend is covered by the new returned window.
    """
    res = []
    for s1, s2 in zip(w1, w2):
        res.append(slice(min(s1.start, s2.start), max(s1.stop, s2.stop)))
    return tuple(res)


""" Module utils/io/files.py 
Contains common operations on creating, checking, listing files/dirs.

API is organized in 3 sections:
1. Post - creation of directories, files
2. Get - query of file types, listing directories or certain files subject
     to name constraints.
"""

import sys, os
import shutil
import pathlib
import ntpath

### Constants ###
IMAGE_EXTS = ['.bmp', '.png', '.jpg', '.jpeg', '.tif', '.tiff', '.gif',
              '.nii', '.nii.gz', '.dcm']


### ======================================================================== ###
### * ### * ### * ### *          API Definitions         * ### * ### * ### * ### 
### ======================================================================== ###

__all__ = ['create_dirs_from_file',
           'is_image', 
           'list_images', 'list_files', 'list_dirs', 'list_all_files_recursive',
           'get_ext', 'get_filename',
           'natural_sort']

### ---- ### ---- \\    POST     // ---- ### ---- ###

def create_dirs_from_file(filepath):
    r""" Creates directory structure (if any part is missing) from filepath. """
    dirs = os.path.dirname(filepath)
    if not dirs:
        return
    if not os.path.exists(dirs):
        os.makedirs(dirs, exist_ok=True)


### ---- ### ---- \\    GET     // ---- ### ---- ###

def is_image(filepath, exts=IMAGE_EXTS, checkfile=False, case_sensitive=False):
    r"""
    Given a file's fullpath, return boolean if it is an image file or not.
    Params:
        filepath - full path to file
        ?exts - list of recognized image extensions (e.g. ['jpg','png'])
    """
    # Doesn't check exts validity for performance reasons
    if checkfile:
        assert os.path.isfile(filepath), f"File ({filepath}) does not exist."
    if isinstance(exts, str):
        exts = [exts]
    assert sum(list(map(lambda x: isinstance(x, str), exts))) == len(exts), \
        f"exts parameter must be a single string or list of strings"
    for ext in exts:
        file_ext = filepath[-len(ext):]
        if (case_sensitive and file_ext == ext) or \
          (not case_sensitive and file_ext.lower() == ext.lower()):
            return True
    return False


def list_images(path, exts=None, sort=True, fullpath=False):
    r"""
    Returns list of images in a given directory with custom specifications.
        Note that list_files is the generalized version of this function.
    Params:
        path - full path to directory of images (asserts existence)
        ?ext - list of custom exts to define image criteria (e.g. ['.jpg','.png])
                if left empty, uses default image extensions
        ?sort - choice to return natural sorted list or not
        ?fullpath - choice to return just image names or full paths
    """
    assert os.path.isdir(path), f"Path ({path}) is not valid."
    if exts is not None:
        assert isinstance(exts, list) or isinstance(exts, tuple)
        assert len(exts) >= 1
    else:
        exts = IMAGE_EXTS
    files_list = [f for f in os.listdir(path)
                  if os.path.isfile(os.path.join(path, f)) and 
                     is_image(os.path.join(path, f), exts=exts)]
    if sort:
        files_list = natural_sort(files_list)
    if fullpath:
        files_list = [os.path.join(path, fn) for fn in files_list]
    return files_list


def list_files(path, subnames=[], hidden=False, sort=True, fullpath=False):
    r"""
    Returns list of files in a directory path with customizable specifications.
    Params:
        path - full path to directory (asserts)
        ?subnames - single str or list of subnames to check for in a file
                    (e.g. ['.jpg','dog'] includes only dog jpg images)
        ?hidden - boolean choice to include hidden files (.*)
        ?sort - boolean choice to natural sort final list
        ?fullpath - boolean choice to return just filenames or full file paths
    """
    if os.path.isfile(path):  # if given file, returns file path with param opts
        if get_filename(path)[0] == '.' and not hidden:
            return []
        if fullpath:
            return [os.path.abspath(path)]
        return [path]
    
    assert os.path.isdir(path), f"Path ({path}) is not valid."
    if isinstance(subnames, str):
        subnames = [subnames]
    assert isinstance(subnames, list)
    assert sum(list(map(lambda x: isinstance(x, str), subnames))) == len(subnames)
    file_list = [f for f in os.listdir(path)
                 if os.path.isfile(os.path.join(path, f))]
    ret_list = []
    for fn in file_list:
        if fn[0] == '.' and not hidden:
            continue
        if not subnames:  # empty list
            ret_list.append(fn)
            continue
        for subname in subnames:
            if subname in fn:
                ret_list.append(fn)
                break 
    if sort:
        ret_list = natural_sort(ret_list)
    if fullpath:
        ret_list = [os.path.abspath(os.path.join(path, fn)) for fn in ret_list]
    return ret_list


def list_dirs(path, subnames=[], hidden=False, sort=True, fullpath=False):
    r"""
    Returns list of directories in a path with customizable specifications.
    Params:
        path - full path to directory (asserts exists) or file
        ?subnames - single str or list of subnames to check for in a directory
                    (e.g. ['data'] includes only names with data in them)
        ?hidden - boolean choice to include hidden directories (.*)
        ?sort - boolean choice to natural sort final list
        ?fullpath - boolean choice to return just dirnames or full dir paths
    """
    assert os.path.isdir(path), f"Path ({path}) is not valid or not a directory."
    if isinstance(subnames, str):
        subnames = [subnames]
    assert isinstance(subnames, list)
    assert sum(list(map(lambda x: isinstance(x, str), subnames))) == len(subnames)
    dir_list = [d for d in os.listdir(path)
                if os.path.isdir(os.path.join(path, d))]
    ret_list = []
    for dn in dir_list:
        if dn[0] == '.' and not hidden:
            continue
        if not subnames:  # empty list
            ret_list.append(dn)
            continue
        for subname in subnames:
            if subname in dn:
                ret_list.append(dn)
                break 
    if sort:
        ret_list = natural_sort(ret_list)
    if fullpath:
        ret_list = [os.path.join(path, dn) for dn in ret_list]
    return ret_list


def list_all_files_recursive(dirpaths, subnames=[], 
                             hidden=False, sort=True, fullpath=True):
    r"""
    Recursively traverses all directories and lists files.
    """
    if isinstance(dirpaths, list) or isinstance(dirpaths, tuple):
        assert sum([os.path.isdir(t) for t in dirpaths]) == len(dirpaths)
    elif isinstance(dirpaths, str):
        assert os.path.isdir(dirpaths), \
               f"Path ({dirpaths}) is not valid or not a directory."
        dirpaths = [dirpaths]
    else:
        raise TypeError(f"Type of dirpaths ({type(dirpaths)}) is invalid.")
    
    files = []
    for dp in dirpaths:
        files += list_files(dp, subnames=subnames, hidden=hidden, sort=sort,
                            fullpath=fullpath)
        more_dirs = list_dirs(dp, hidden=hidden, fullpath=True)
        for mdp in more_dirs:
            files += list_all_files_recursive(mdp, subnames=subnames)

    return files


def get_ext(path, dot=True):
    r"""
    Returns file extension given a single filename or its full path.
    Params:
        path - full path to file (does not check if file actually exists)
        ?dot - choice of whether to return extension with dot in front or not
    """
    cleaned_fn = get_filename(path, ext=True)
    _, dot_ext = os.path.splitext(cleaned_fn)
    if dot:
        return dot_ext
    return dot_ext[1:]


def get_filename(path, ext=True):
    r"""
    Returns file name given a single filename or its full path.
    Params:
        path - full path to file (does not check if file actually exists)
        ?ext - choice of whether to return filename with an extension or not
    """
    # with paths ending in '/', removes that and assigns to head with tail=None
    head, tail = ntpath.split(path)  
    fn = tail or ntpath.basename(head)
    if ext:
        return fn
    return fn[:len(fn)-len(get_ext(fn, dot=True))]


### ---- ### ---- \\    DELETE     // ---- ### ---- ###


def delete_file(filepath):
    r""" Deletes a file if it exists. """
    if os.path.isfile(filepath):
        os.remove(filepath)


### ---- ### ---- \\    Other Utilities     // ---- ### ---- ###

def natural_sort(name_list):
    r"""
    Returns new, sorted list of name_list.
    """
    assert isinstance(name_list, list)
    if len(name_list) == 0:
        return []
    assert sum(list(map(lambda x: isinstance(x, str), name_list))) == len(name_list)
    
    from re import compile, split
    dre = compile(r'(\d+)')
    ret_list = sorted(name_list, 
                      key=lambda l: [int(s) if s.isdigit() else s.lower() 
                                     for s in split(dre, l)])
    return ret_list



### ======================================================================== ###
### * ### * ### * ### *         Rudimentary Tests        * ### * ### * ### * ### 
### ======================================================================== ###    


if __name__ == '__main__':
    
    # TEST is_image
    print(f"\n--------\nTesting list_files..\n")
    img = '/subset_train/0_0/1.png'
    # print(f"img ({img}) --> {is_image(img, exts=['jpg'])}")
    img = '/subset_train/0_0/1.png'
    print(f"img ({img}) --> {is_image(img, exts=[])}")
    img = '/subset_train/0_0/1.png'
    print(f"img ({img}) --> {is_image(img, exts=['.png'])}")

    # TEST list_files
    print(f"\n--------\nTesting list_files..\n")
    ex_dir = '/Users/charzhar/Desktop/[Project] rere/rere/datasets/data/mnist/subset_train/0_0'
    print(list_files('/Users/charzhar/Desktop/[Project] rere/rere/datasets/data/others'))
    print(list_files(ex_dir, subnames=['tst']))
    print(list_files(ex_dir, subnames=['111','2222']))
    print(list_files(ex_dir, subnames=['599'], fullpath=True)[:5], '...')

    # TEST list_dirs
    print(f"\n--------\nTesting list_dirs..\n")
    ex_dir = '/Users/charzhar/Desktop/[Project] rere/rere/datasets/data/mnist/subset_train'
    print(list_dirs('/Users/charzhar/Desktop/[Project] rere/rere/datasets/data/mnist/subset_train/0_0'))
    print(list_dirs(ex_dir, subnames=['A','1','2']))
    print(list_dirs(ex_dir, subnames=['A','1','2','_']))
    print(list_dirs(ex_dir, subnames=['1','2','3'], fullpath=True))

    # TEST get_filename and get_ext
    print()
    fn = 'mn/ist01.png'
    print(f"ext for ({fn}) -> ({get_ext(fn)})")
    fn = 'mn.ist01.png'
    print(f"ext for ({fn}) -> ({get_ext(fn)})")
    fn = '/desktop/mnistnotes'
    print(f"ext for ({fn}) -> ({get_ext(fn)})")
    fn = '/desktop/mnistnotes.'
    print(f"ext for ({fn}) -> ({get_ext(fn)})")
    fn = '/desktop/mnistnotes./'
    print(f"ext for ({fn}) -> ({get_ext(fn)})")
    fn = '/desktop/mn.ist01.png/'
    print(f"ext for ({fn}) -> ({get_ext(fn)})")
    fn = '/desktop/.mn.ist01.png'
    print(f"ext for ({fn}) -> ({get_ext(fn, dot=False)})")
    
    print()
    fn = 'mnist01.png'
    print(f"filename for ({fn}) -> ({get_filename(fn)})")
    fn = 'mnist01.png/'
    print(f"filename for ({fn}) -> ({get_filename(fn)})")
    fn = '/mnist01.png'
    print(f"filename for ({fn}) -> ({get_filename(fn)})")
    fn = '/mnist01.png/'
    print(f"filename for ({fn}) -> ({get_filename(fn)})")
    fn = 'desktop/mnist/mnist01.png/'
    print(f"filename for ({fn}) -> ({get_filename(fn, ext=False)})")
    fn = 'desktop/mnist/mnistnotes/'
    print(f"filename for ({fn}) -> ({get_filename(fn, ext=False)})")




""" Module basemodel.py (By: Charley Zhang, July 2020)
Implements basic functionality for all models subclasses.
"""

import sys, os
import torch, torch.nn as nn
import torchsummary


class BaseModel(nn.Module):
    r""" Pytorch basemodel with useful customized functionalities."""
    def __init__(self, *args, **kwargs):
        super(BaseModel, self).__init__(*args, **kwargs)

    def forward(self, *args):
        raise NotImplementedError(f"forward() func requires definition.")

    @property
    def device(self):
        return next(self.parameters()).device if self.parameters() else None

    @property
    def param_counts(self):
        tot_params = sum(p.numel() for p in self.parameters())
        tot_train_params = sum(p.numel() for p in self.parameters()
            if p.requires_grad
        )
        return tot_params, tot_train_params

    @property
    def size(self):
        r""" Gets total parameter and buffer memory usage in bytes. """
        params_mem = sum(
            [p.nelement() * p.element_size() for p in self.parameters()]
        )
        bufs_mem = sum(
            [buf.nelement() * buf.element_size() for buf in self.buffers()]
        )
        return params_mem + bufs_mem 

    def summary(self, input_size=(3,256,256), batch_size=-1, device='cpu'):
        if 'cuda' in device:
            device = 'cuda'  # summary does not support targ device assignment
        torchsummary.summary(
            self, 
            input_size=input_size, 
            batch_size=batch_size,
            device=device
        )

        
    


""" Module metrics3d.py (By: Charley Zhang, 2021.05)
Metrics for 3D segmentation. 

File contains 3 sections:
(1) Grouped metrics - func that return mutliple metrics (optimized execution)
(2) Individual metrics 
(3) Helper utilities for 3D+ tensor/array operations

Notes of Usage & Assumptions:
 - Assumes 3D input (X, Y, Z), 2D metrics should be in metrics2d.py
 - Assumes torch-like channel-first view BxCxHxWxD or CxHxWxD
 - Assumes no differentiability requirements.
 - Top-level func 'standard_metrics' should be sufficient for most uses.
   This function takes in tensors or np arrays. 
   
Resources: 
 - https://github.com/JunMa11/SegLoss/blob/master/losses_pytorch/dice_loss.py
"""


import torch
import numpy as np



def standard_metrics(pred, targ, get_hausdorff=False):
    """ Returns dictionary of most commonly used segmentation metrics.
    Parameters
        pred - prediction matrix BxCxHxWxD
        targ - target gt matrix  Bx1xHxWxD
    Return 
      - Dict: element x class-wise (BxC) TP, TN, FP, FN,
        Class-wise(Cx) dice ceofficient, jaccard (IOU), and Hausdorff.
    """
    pred, targ = standardize_dims(pred, targ)
    tp, fp, fn = confusion_matrix(pred, targ)  # BxC matrix
    
    # class-wise dice loss
    tp_c, fp_c, fn_c = tp.sum(0), fp.sum(0), fn.sum(0)
    dice_c = 2 * tp_c / (2 * tp_c + fp_c + fn_c + 10**-7)
    jaccard_c = tp_c / (tp_c + fp_c + fn_c + 10**-7)
    recall_c = tp_c / (tp_c + fn_c + 10**-7)
    
    hausd_c = hausdorff(pred, targ) if get_hausdorff else None
    
    return {
        'tp': tp,  # TPs as BxC matrix
        'fp': fp,
        'fn': fn,
        'tp_c': tp_c, # _c stands for per class (a size C vector) 
        'fp_c': fp_c,
        'fn_c': fn_c,
        'dice_c': dice_c,  
        'jaccard_c': jaccard_c,
        'hausdorff_c': hausd_c,
        'sensitivity_c': recall_c,
    }


### ======================================================================== ###
### * ### * ### * ### *      3D Segmentation Metrics     * ### * ### * ### * ###
### ======================================================================== ###


def confusion_matrix(pred, targ):
    """ Return counts for TPs, FPs, FN, TNs. 
    Parameters
        pred - prediction tensor or np array [Bx[Cx]]HxWxD
        targ - target tensor or np array  [Bx[Cx,1x]]HxWxD
    Return
        Dict of tp, fp, fn values per instance and class BxC.
          ex. tp is BxC array where with bth row & cth col is the # of
            true positives for example b in the batch and class c. 
    """
    pred, targ = standardize_dims(pred, targ)
    axes = tuple(range(2, pred.ndim))

    tp = pred * targ
    tp = sum_tensor(tp, axes, keepdim=False)
    # import IPython; IPython.emb ed(); 
    fp = pred * (1 - targ)
    fp = sum_tensor(fp, axes, keepdim=False)
    fn = (1 - pred) * targ
    fn = sum_tensor(fn, axes, keepdim=False)
    
    return tp, fp, fn


def hausdorff(pred, targ):
    pass


### ======================================================================== ###
### * ### * ### * ### *          Helper Utilities        * ### * ### * ### * ###
### ======================================================================== ###


def sum_tensor(inp, axes, keepdim=False):
    """ 
    https://github.com/MIC-DKFZ/nnUNet/blob/master/nnunet/utilities/tensor_utilities.py
    """
    axes = np.unique(tuple(axes)).astype(int)
    if keepdim:
        for ax in axes:
            inp = inp.sum(int(ax), keepdim=True)
    else:
        for ax in sorted(axes, reverse=True):
            inp = inp.sum(int(ax))
    return inp


def np_scatter(slf, dim, index, src):
    """
    Taken from:
        https://stackoverflow.com/questions/46065873/how-to-do-scatter-and-gather-operations-in-numpy
    Writes all values from the Tensor src into new matrix at the indices 
    specified in the index Tensor.
    Example:
        slf[index[i][j][k]][j][k] = src[i][j][k]  # if dim == 0
        slf[i][index[i][j][k]][k] = src[i][j][k]  # if dim == 1
        slf[i][j][index[i][j][k]] = src[i][j][k]  # if dim == 2
    Paramters
        slf - Target matrix to which values will be placed
        dim - The axis along which to index
        index - The indices of elements to scatter
        src: The source element(s) to scatter
    Return
        slf - Matrix with changed values
    """
    if index.dtype != np.dtype('int_'):
        raise TypeError("The values of index must be integers")
    if slf.ndim != index.ndim:
        raise ValueError("Index should have the same number of dimensions as output")
    if dim >= slf.ndim or dim < -slf.ndim:
        raise IndexError("dim is out of range")
    if dim < 0:
        # Behavior in PyTorch's scatter where dim < 0 
        dim = slf.ndim + dim
    idx_xsection_shape = index.shape[:dim] + index.shape[dim + 1:]
    slf_xsection_shape = slf.shape[:dim] + slf.shape[dim + 1:]
    if idx_xsection_shape != slf_xsection_shape:
        raise ValueError("Except for dimension " + str(dim) +
                         ", all dimensions of index and output " + \
                         "should be the same size")
    if (index >= slf.shape[dim]).any() or (index < 0).any():
        raise IndexError("The values of index must be between 0 "+ \
                         "and (slf.shape[dim] -1)")

    def make_slice(arr, dim, i):
        slc = [slice(None)] * arr.ndim
        slc[dim] = i
        return slc

    # We use index and dim parameters to create idx
    #  idx is in a form that can be used as a NumPy advanced index for 
    #  scattering of src param. in self
    # import IPython; IPython.embed(); 
    idx = [[*np.indices(tuple(idx_xsection_shape)).reshape(index.ndim - 1, -1),
            index[tuple(make_slice(index, dim, i))].reshape(1, -1)[0]] \
            for i in range(index.shape[dim])]
    idx = list(np.concatenate(idx, axis=1))
    idx.insert(dim, idx.pop())

    if not np.isscalar(src):
        if index.shape[dim] > src.shape[dim]:
            raise IndexError("Dimension " + str(dim) + \
                             "of index can not be bigger than that of src ")
        src_xsection_shape = src.shape[:dim] + src.shape[dim + 1:]
        if idx_xsection_shape != src_xsection_shape:
            raise ValueError("Except for dimension " +
                             str(dim) + ", all dimensions of index and src " + \
                             "should be the same size")
        # src_idx is a NumPy advanced index for indexing of elements in the src
        src_idx = list(idx)
        src_idx.pop(dim)
        src_idx.insert(dim, np.repeat(np.arange(index.shape[dim]), 
                                      np.prod(idx_xsection_shape)))
        slf[idx] = src[src_idx]

    else:
        slf[tuple(idx)] = src
    return slf


def scatter(matrix, dim, index, src):
    if isinstance(matrix, torch.Tensor):
        return matrix.scatter_(dim, index.to(torch.int64), src)
    else:
        return np_scatter(matrix, dim, index, src)
        

def unsqueeze(matrix, dim=0):
    """ Unsqueeze a tensor or ndarray on dim. """
    if isinstance(matrix, torch.Tensor):
        return matrix.unsqueeze(dim)
    else: 
        return np.expand_dims(matrix, dim)


def standardize_dims(pred, targ):
    """ Standardize pred and targ to be BxCxHxWxD & BxCxHxWxD, resp.
    Parameters
        pred - can be HxWxD, CxHxWxD, or BxCxHxWxD (assume C to be # classes)
        targ - can be HxWxD, CxHxWxD, or BxCxHxWxD (C = N_C or 1)
    """
    assert pred.ndim in (3, 4, 5), (f"Invalid prediction shape {pshape}. "
        f"Prediction must be HxWxD, CxHxWxD, or CxHxWxD")
    assert targ.ndim in (3, 4, 5), (f"Invalid target shape {tshape}. "
        f"Prediction must be HxWxD, CxHxWxD, or CxHxWxD")
    assert pred.shape[-3:] == targ.shape[-3:], (f"Pred, Targ size mismatch! "
        f"Pred: {pred.shape[-3:]}, Targ: {targ.shape[-3:]}")
    
    with torch.no_grad():  # improve performance in case inputs are tensors
        # standardize prediction matrix
        if pred.ndim == 3:   # HxWxD
            std_pred = unsqueeze(unsqueeze(pred))
        elif pred.ndim == 4: # CxHxWxD
            std_pred = unsqueeze(pred)
        else:
            std_pred = pred
        
        # standardize target matrix to BxCxHxWxD one-hot
        C = std_pred.shape[1]
        # print(std_pred.shape)
        if targ.ndim == 3:   # HxWxD
            std_targ = unsqueeze(unsqueeze(targ))
            return std_pred, std_targ
        
        if targ.ndim == 4:        
            if targ.shape[0] == C:  
                return std_pred, unsqueeze(targ)
            if targ.shape[0] == 1:
                targ = unsqueeze(targ)
            else:
                raise f"Targ neither matches pred channels nor has 1 channel."
        
        # ndim = 5
        if targ.shape[1] == std_pred.shape[1]:
            return std_pred, targ
        
        # make into 1-hot from labels of single target channel
        if isinstance(targ, torch.Tensor):
            std_targ = torch.zeros(std_pred.shape, device=targ.device)
        else:
            std_targ = np.zeros(std_pred.shape)
        
        return std_pred, scatter(std_targ, 1, targ, 1)


### ======================================================================== ###
### * ### * ### * ### *                Tests             * ### * ### * ### * ###
### ======================================================================== ###


if __name__ == '__main__':
    
    ### Standardize Dims ###
    C = 2  # num_channels
    for pshape, tshape in [[(1, 2, 3),(1, 2, 3)], [(2, 1, 2, 3),(2, 1, 2, 3)],
                [(2, 1, 2, 3),(1, 1, 2, 3)], [(3, 2, 1, 2, 3),(3, 2, 1, 2, 3)],
                [(3, 2, 1, 2, 3),(3, 1, 1, 2, 3)]]:
        print(f"Testing shapes: pred{pshape}, targ{tshape}..", end='')
        pred_tens, targ_tens = torch.randn(pshape), torch.randint(0, C, tshape)
        p, t = standardize_dims(pred_tens, targ_tens)
        assert p.shape == t.shape
        assert p.ndim == 5
        
        pred_np = pred_tens.cpu().numpy()
        targ_np = targ_tens.cpu().numpy()
        p_np, t_np = standardize_dims(pred_np, targ_np)
        assert p_np.shape == t_np.shape
        assert p_np.ndim == 5
        
        assert (p.cpu().numpy() == p_np).all()
        assert (t.cpu().numpy() == t_np).all()
        
        print('..done ✔')
    
    ### Standard Metrics ###
    print(f"\n----------\nTesting Standard Metrics")
    pred = torch.randint(0, 2, (3, 2, 1, 2, 3))
    targ = torch.randint(0, 2, (3, 2, 1, 2, 3))
    print('Prediction:\n', pred, '\nTarget:\n', targ)
    
    d = standard_metrics(pred, targ)
    
    rtp, rfp, rfn = d['tp'].sum(), d['fp'].sum(), d['fn'].sum()
    tp = ((pred == 1) & (targ == 1)).sum()
    fp = ((pred == 1) & (targ == 0)).sum()
    fn = ((pred == 0) & (targ == 1)).sum()
    assert tp == rtp, f"{tp} {rtp}"
    assert fp == rfp, f"{fp} {rfp}"
    assert fn == rfn, f"{fn} {rfn}"
    print(f"Correct ✔")




import pathlib
from setuptools import setup, find_packages

current_dir = pathlib.Path(__file__).parent
readme = (current_dir / "README.md").read_text()
requirements = [
    'SimpleITK!=2.0.*',  # https://github.com/SimpleITK/SimpleITK/issues/1239
    'nibabel',
    'numpy>=1.15',
    'scipy',
    'scikit-image',
    'Pillow',
    'torch>=1.1',
    'torchvision',
    'torch-summary',
    'pandas',
    'psutil',
    'matplotlib',
    'wandb',
    'tqdm',
]

setup(
    name="deep-medical-toolkit",
    keywords='dmt',
    version="0.0.4",
    description=("Tools to facilitate deep learning research with a focus on "
    			 "medical imaging."),
    long_description=readme,
    long_description_content_type="text/markdown",
    # url="TBA",
    author="Charley Zhang",
    author_email="yzhang46@nd.edu",
    license="Apache License 2.0",
    classifiers=[
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(include=["dmt", "dmt.*"]),
    include_package_data=True,
    install_requires=requirements,
)





import SimpleITK as sitk
import numpy as np
import torch


### ---- ### ---- \\    Keys     // ---- ### ---- ###

IMAGE = 'dmt_default_key_image'
SHAPE = 'dmt_default_key_shape'
LOCATION = 'dmt_default_key_location'  # location = lower_indices + upper_ind

### ---- ### ---- \\    Image Constants     // ---- ### ---- ###


IMTYPE_2_NP = {
    'bool': np.bool_,
    'uint8': np.uint8,
    'uint16': np.uint16,
    'uint32': np.uint32,
    'int8': np.int8,
    'int16': np.int16,
    'int': np.int32,
    'int32': np.int32,
    'long': np.int64,
    'int64': np.int64,
    'float': np.float32,
    'float32': np.float32,
    'double': np.float64,
    'float64': np.float64
}

IMTYPE_2_TORCH = {
    'bool': torch.bool,
    'uint8': torch.uint8,
    'int8': torch.int8,
    'int16': torch.int16,
    'int': torch.int32,
    'int32': torch.int32,
    'long': torch.int64,
    'int64': torch.int64,
    'float': torch.float32,
    'float32': torch.float32,
    'double': torch.float64,
    'float64': torch.float64
}

IMTYPE_2_SCALAR_SITK = {
    'uint8': sitk.sitkUInt8,
    'uint16': sitk.sitkUInt16,
    'uint32': sitk.sitkUInt32,
    'int8': sitk.sitkInt8,
    'int16': sitk.sitkInt16,
    'int': sitk.sitkInt32,
    'int32': sitk.sitkInt32,
    'long': sitk.sitkInt64,
    'int64': sitk.sitkInt64,
    'float': sitk.sitkFloat32,
    'float32': sitk.sitkFloat32,
    'double': sitk.sitkFloat64,
    'float64': sitk.sitkFloat64
}

IMTYPE_2_VECTOR_SITK = {
	'uint8': sitk.sitkVectorUInt8,
    'uint16': sitk.sitkVectorUInt16,
    'uint32': sitk.sitkVectorUInt32,
    'int8': sitk.sitkVectorInt8,
    'int16': sitk.sitkVectorInt16,
    'int': sitk.sitkVectorInt32,
    'int32': sitk.sitkVectorInt32,
    'long': sitk.sitkVectorInt64,
    'int64': sitk.sitkVectorInt64,
    'float': sitk.sitkVectorFloat32,
    'float32': sitk.sitkVectorFloat32,
    'double': sitk.sitkVectorFloat64,
    'float64': sitk.sitkVectorFloat64
}

IMAGE_TYPES_STR = ['bool', 'uint8', 'uint16', 'uint32', 'int8', 'int16', 
               'int', 'int32', 'long', 'int64', 
               'float', 'float32', 'double', 'float64']

IMAGE_TYPES = list(IMTYPE_2_NP.values()) + list(IMTYPE_2_TORCH.values()) + \
    list(IMTYPE_2_SCALAR_SITK.values()) + list(IMTYPE_2_VECTOR_SITK.values())
    
INTERPOLATIONS_2_SITK_INTERPOLATIONS = {
    'nearest': 'sitkNearestNeighbor',  # Copies NN position of non-int pixel position
    'linear': 'sitkLinear', # Linearly interpolates intensity at a non-integer pixel position.
    'bspline': 'sitkBSpline',  # Cubic interpolation
    'cubic': 'sitkBSpline',  # Same as bspline
    'gaussian': 'sitkGaussian',  # sigma=0.8, alpha=4
    'label_gaussian': 'sitkLabelGaussian',  # smooth interp multi-label images, σ=1, α=1
    'hamming': 'sitkHammingWindowedSinc',  # Hamming windowed sinc kernel.
    'cosine': 'sitkCosineWindowedSinc',  # Cosine windowed sinc kernel.
    'welch': 'sitkWelchWindowedSinc',  # Welch windowed sinc kernel.
    'lanczos': 'sitkLanczosWindowedSinc',  #: Lanczos windowed sinc kernel.
    'blackman': 'sitkBlackmanWindowedSinc'  #: Blackman windowed sinc kernel.
}


""" schedulers.py (By: Charley Zhang, June 2020)
Various effective learning rate schedulers taken from literature.
"""

import sys
import math


class BaseScheduler:
    def step(self, epoch=-1, val=-1):
        raise NotImplementedError()
    
    def set_lr(self, lr):
        """(1) set all optim param_groups to lr; (2) update self.lr"""
        for g in self.optimizer.param_groups: # list of dicts with 'lr', 'params'
            g['lr'] = lr
        self.lr = self.get_lr()

    def get_lr(self):
        return next(iter(self.optimizer.param_groups))['lr']


class Uniform(BaseScheduler):
    """ Keeps uniform learning rate throughout. """
    def __init__(self, optimizer, rampup_rates=[]):
        self.optimizer = optimizer
        self.lr = self.get_lr()
        self.rampup_rates = rampup_rates
        if self.rampup_rates:
            if self.rampup_rates[-1] != self.lr:
                self.rampup_rates.append(self.lr)
            self.set_lr(self.rampup_rates.pop(0))

    def step(self, **kws):
        if self.rampup_rates:
            self.set_lr(self.rampup_rates.pop(0))


class ExponentialDecay(BaseScheduler):
    """ LR_t = orig_lr * (exp_fac) ^ t """
    def __init__(self, optimizer, exp_factor=0.9, t=0, minlr=1e-8, rampup_rates=[]):
        self.exp_factor = exp_factor
        self.minlr = minlr
        self.optimizer = optimizer
        self.lr = self.orig_lr = self.get_lr()

        self.rampup_rates = rampup_rates
        self.step(epoch=t)   # in-case we start from epoch > 1

    def step(self, epoch=None, **kws):
        assert epoch is not None, "Need to give epoch."
        
        if self.rampup_rates:
            self.set_lr(self.rampup_rates.pop(0))
        else:
            new_lr = max(self.minlr, self.orig_lr * (self.exp_factor) ** epoch)
            if new_lr != self.lr:
                self.set_lr(new_lr)


class LinearDecay(BaseScheduler):
    """ LR_t = LR_t+1 - (end_factor / T)  """
    def __init__(self, optimizer, end_factor=0.01, T=60, minlr=1e-8, 
                 rampup_rates=[]):
        self.end_factor = end_factor
        self.minlr = minlr
        self.optimizer = optimizer
        self.lr = self.orig_lr = self.get_lr()
        self.rampup_rates = rampup_rates

        self.step_reduction = (self.lr - self.lr * self.end_factor) / T
        print(f"\t (LinearDecay) Scheduler initialized with final lr={self.lr * self.end_factor:.8f}.")

    def step(self, **kws):
        if self.rampup_rates:
            self.set_lr(self.rampup_rates.pop(0))
        else:
            new_lr = max(self.minlr, self.lr - self.step_reduction)
            if new_lr != self.lr:
                self.set_lr(new_lr)


class ConsistencyCosineDecay(BaseScheduler):
    """ Scheduler used commonly in consistency regularization methods.
    LR_t = min_lr + orig_lr * cos(7*pi*t/(16*T)) 
    """
    def __init__(self, optimizer, T, t=0, minlr=1e-8, rampup_rates=[]):
        self.T = T
        self.minlr = minlr
        self.optimizer = optimizer
        self.lr = self.orig_lr = self.get_lr()

        self.rampup_rates = rampup_rates
        self.step(epoch=t)   # in-case we start from epoch > 1

    def step(self, epoch=None, **kws):
        assert epoch is not None, "Need to give epoch."
        
        if self.rampup_rates:
            self.set_lr(self.rampup_rates.pop(0))
        else:
            new_lr = max(self.minlr, self.orig_lr * math.cos(
                            7 * math.pi * epoch / (16 * self.T)))
            if new_lr != self.lr:
                self.set_lr(new_lr)


class CosineDecay(BaseScheduler):
    """ LR_t = min_lr + 0.5(1 + cos(pi*t/T)) * LR_0 """
    def __init__(self, optimizer, T, t=0, minlr=1e-8, rampup_rates=[]):
        self.T = T
        self.minlr = minlr
        self.optimizer = optimizer
        self.lr = self.orig_lr = self.get_lr()

        self.rampup_rates = rampup_rates
        self.step(epoch=t)   # in-case we start from epoch > 1

    def step(self, epoch=None, **kws):
        assert epoch is not None, "Need to give epoch."
        
        if self.rampup_rates:
            self.set_lr(self.rampup_rates.pop(0))
        else:
            new_lr = self.minlr + 0.5 * (1 + math.cos(
                math.pi * epoch / self.T)) * self.orig_lr
            if new_lr != self.lr:
                self.set_lr(new_lr)

    
class StepDecay(BaseScheduler):
    
    def __init__(self, optimizer, T, factor=0.2, steps=[0.33, 0.66], 
            rampup_rates=[]):
        self.T = T
        self.optimizer = optimizer
        self.factor = factor
        self.lr = self.start_lr = self.get_lr()

        if not steps or not sum(steps):
            self.step_epochs = []
        else:
            mult_fac = T if steps[0] < 1 else 1
            self.step_epochs = [int(s * mult_fac) for s in steps]

        self.rampup_rates = rampup_rates
        if self.rampup_rates:
            if self.rampup_rates[-1] != self.lr:
                self.rampup_rates.append(self.lr)
            self.set_lr(self.rampup_rates.pop(0))

    def step(self, epoch=None, **kws):
        assert epoch is not None
        if self.rampup_rates:
            self.set_lr(self.rampup_rates.pop(0))
            return

        num_lrsteps = sum([epoch >= s for s in self.step_epochs])
        correct_lr = self.start_lr * self.factor ** num_lrsteps
        if self.lr != correct_lr:
            self.set_lr(correct_lr)


class ReduceOnPlateau(BaseScheduler):
    """ Reduce LR on Plateue scheduler based on loss."""
    
    def __init__(self, optimizer, factor=0.5, patience=3, lowerbetter=True,
            rampup_rates=[]):
        self.factor = 0.5
        self.patience = patience
        self.lowerbetter = lowerbetter
        
        self.best_val = sys.float_info.max if lowerbetter else -sys.float_info.max
        self.bad_iter_count = 0

        self.optimizer = optimizer
        self.lr = self.get_lr()

        self.rampup_rates = rampup_rates
        if self.rampup_rates:
            if self.rampup_rates[-1] != self.lr:
                self.rampup_rates.append(self.lr)
            self.set_lr(self.rampup_rates.pop(0))
    
    def step(self, val=None, **kws):
        assert val is not None, "Need to give step a val."

        if self.rampup_rates:
            self.set_lr(self.rampup_rates.pop(0))
            return

        if self.lowerbetter:
            improved = True if val <= self.best_val else False
        else: 
            improved = True if val >= self.best_val else False
        
        if improved:
            self.best_val = val
            self.bad_iter_count = 0
        else:
            self.bad_iter_count += 1
            if self.bad_iter_count >= self.patience:
                self.set_lr(self.lr * self.factor)
                self.bad_iter_count = 0
                self.best_val = val
        

# Rudimentary Tests
if __name__ == '__main__':

    import torch

    T=20
    LR=1.
    FACTOR=0.5

    def get_optim():
        net = torch.nn.Sequential(
            torch.nn.Conv2d(12, 24, kernel_size=(3,3), stride=1, padding=1)
        )
        return torch.optim.SGD(net.parameters(), lr=LR)
    
    # stand-alone tests
    optimizer = get_optim()
    step = StepDecay(optimizer, T, factor=FACTOR, steps=[5,10,15,25])
    assert step.step_epochs == [5,10,15,25], f'{step.step_epochs}'
    step = StepDecay(optimizer, T, factor=FACTOR, steps=[.5,.8,2.])
    assert step.step_epochs == [10,16,40], f'{step.step_epochs}'
    cos = CosineDecay(optimizer, T, t=0)
    assert cos.lr == 1., f'{cos.lr}'
    cos = CosineDecay(optimizer, T, t=5)
    assert cos.lr == 0.5*(1+math.cos(math.pi*5/T))*LR, f'{cos.lr}'

    # start epoch = 0
    optimizer = get_optim()
    cos = CosineDecay(optimizer, T, t=0)
    step = StepDecay(optimizer, T, factor=FACTOR, steps=[5,10,15,25])
    plat_l = ReduceOnPlateau(optimizer, factor=FACTOR, patience=5, lowerbetter=True)
    plat_h = ReduceOnPlateau(optimizer, factor=FACTOR, patience=5, lowerbetter=False)
    
    schedulers = [cos, step, plat_l, plat_h]
    lrs = []
    for epoch in range(1, T+1):
        [s.step(epoch=epoch, val=epoch) for s in schedulers]
        lrs.append([s.lr for s in schedulers])

    cos_ans = [0.5 * (1 + math.cos(math.pi*e/T)) * LR for e in range(1, T+1)]
    step_ans = [LR]*4 + [LR*FACTOR]*5 + [LR*FACTOR**2]*5 + [LR*FACTOR**3]*6
    plat_lower_ans = [LR]*5 + [LR*FACTOR]*5 + [LR*FACTOR**2]*5 + [LR*FACTOR**3]*5
    plat_higher_ans = [LR]*20
    ans = [cos_ans, step_ans, plat_lower_ans, plat_higher_ans]

    for i, l in enumerate(zip(*lrs)):
        print(f"Testing schedule #{i}")
        assert ans[i] == list(l), f"\nAns{ans[i]}\nOut{l}"

    # start epoc_h = 5
    optimizer = get_optim()
    cos = CosineDecay(optimizer, T, t=5)
    optimizer = get_optim()
    step = StepDecay(optimizer, T, factor=FACTOR, steps=[5,10,15,25])
    plat_l = ReduceOnPlateau(optimizer, factor=FACTOR, patience=5, lowerbetter=True)
    plat_h = ReduceOnPlateau(optimizer, factor=FACTOR, patience=5, lowerbetter=False)
    
    schedulers = [cos, step, plat_l, plat_h]
    lrs = []
    for epoch in range(5, T+1):
        [s.step(epoch=epoch, val=epoch) for s in schedulers]
        lrs.append([s.lr for s in schedulers])

    cos_ans = [0.5 * (1 + math.cos(math.pi*e/T)) * LR for e in range(5, T+1)]
    step_ans = [LR*FACTOR]*5 + [LR*FACTOR**2]*5 + [LR*FACTOR**3]*6
    plat_lower_ans = [LR]*5 + [LR*FACTOR**1]*5 + [LR*FACTOR**2]*5 + [LR*FACTOR**3]
    plat_higher_ans = [LR]*16
    ans = [cos_ans, step_ans, plat_lower_ans, plat_higher_ans]

    for i, l in enumerate(zip(*lrs)):
        print(f"Testing schedule #{i}")
        assert ans[i] == list(l), f"\nAns{ans[i]}\nOut{l}"
    
    
    print("✔ All passed ✔")

    


""" dmt/data/loading/mtm_loader.py
Multi-sample to multi-example multiprocessed loader. 
Multi-sample means that it requires multiple (>=1) dataset samples to create
multiple (>=1) batch examples. It is a generalized form of OTO & OTM loaders.

TODO: when project demands it, implement & add to data __init__ file.
"""

class ManyToManyLoader:
    pass